// Unreal Fighting Framework by Kevin Poretti

#include "FFStateMachineComponent.h"

// FF includes
#include "FFState.h"

UFFStateMachineComponent::UFFStateMachineComponent()
{
    // Don't use Unreal's tick instead use a fixed tick
	PrimaryComponentTick.bCanEverTick = false;
}


void UFFStateMachineComponent::Initialize()
{
    for(const TSubclassOf<UFFState>& CurrState : DefaultStates)
    {
        UFFState* TempState = AddState(CurrState);
        if(!CurrentState) // first state to be created is the entry into this state machine
        {
            CurrentState = TempState;
            CurrentState->Enter(GetCurrentStateContext());
        }
    }
}


void UFFStateMachineComponent::InitActorInfo(AActor* InOwner, AActor* InAvatar)
{
    Owner = InOwner;
	Avatar = InAvatar;
}


UFFState* UFFStateMachineComponent::AddState(TSubclassOf<UFFState> StateClassToAdd)
{
    UFFState* TempState = NewObject<UFFState>(this, StateClassToAdd);
    if(TempState)
    {
        States.Add(TempState);
        return TempState;
    }

    return nullptr;
}


void UFFStateMachineComponent::AddStates(const TArray<TSubclassOf<UFFState>>& StateClassesToAdd)
{
    for(const TSubclassOf<UFFState>& CurrState : StateClassesToAdd)
    {
        AddState(CurrState);
    }
}


void UFFStateMachineComponent::RemoveState(FName StateToRemove)
{
    UE_LOG(LogTemp, Error, TEXT("UFFStateMachineComponent::RemoveState is not yet implemented"));
}

void UFFStateMachineComponent::SwitchStates(UFFState* NewState)
{
    check(CurrentState);
    check(NewState);

    CurrentState->Exit(GetCurrentStateContext());
    CurrentState = NewState;
    CurrentState->Enter(GetCurrentStateContext());
    TicksInState = 0;
    CurrentSubStateLabel = NAME_None;
}


FName UFFStateMachineComponent::GetCurrentStateName() const
{
    check(CurrentState)

    return CurrentState->Name;
}


FFFStateContext UFFStateMachineComponent::GetCurrentStateContext() const
{
    FFFStateContext CurrStateContext;
    CurrStateContext.Owner = Owner;
    CurrStateContext.Avatar = Avatar;
    CurrStateContext.Parent = this;
    CurrStateContext.TicksInState = TicksInState;
    return CurrStateContext;
}


void UFFStateMachineComponent::SetSubStateLabel(FName InSubStateLabel)
{
    CurrentSubStateLabel = InSubStateLabel;
}


UFFState* UFFStateMachineComponent::FindStateWithName(FName StateName)
{
    for (UFFState* CurrState : States)
    {
        if(CurrState->Name == StateName)
        {
            return CurrState;
        }
    }

	UE_LOG(LogTemp, Warning,
			TEXT("Could not find state in state machine with name %s on %s"), *StateName.ToString(), *Owner->GetName());

	return nullptr;
}


void UFFStateMachineComponent::BeginPlay()
{
	Super::BeginPlay();

    Initialize();
}


void UFFStateMachineComponent::FixedTick(float OneFrame)
{
    // CurrentState should never be null
    // TODO: Should probably assert or whatever UE's equivalent is
    check(CurrentState);

    // Should we switch states?
    UFFState* StateToTransitionTo = nullptr;
    for(UFFState* ThisState : States)
    {
        // found a state 
        if(ThisState->CanTransition(GetCurrentStateContext()))
        {
            StateToTransitionTo = ThisState;
            break;
        }
    }

    // Lastly just check if the state we're about to transition into isn't the current state.
    // It is OK to transition if state's "CanTransitionToSelf" is true
    if(StateToTransitionTo && 
        (CurrentState->Name != StateToTransitionTo->Name || StateToTransitionTo->bCanTransitionToSelf))
    {
        SwitchStates(StateToTransitionTo);
    }
    else
    {
        // Tick current state
        TicksInState++;
        CurrentState->Update(OneFrame, GetCurrentStateContext());   
    }

    // Debug
}

